@@ -13,19 +13,25 @@
 #
 # Copyright Buildbot Team Members
 
+from __future__ import absolute_import
+from __future__ import print_function
+
 import os
 
-from twisted.python import log
 from twisted.internet import defer
+from twisted.internet import reactor
+from twisted.python import log
 
+from buildbot.interfaces import WorkerTooOldError
 from buildbot.process import buildstep
+from buildbot.process import remotecommand
 from buildbot.steps.source.base import Source
-from buildbot.interfaces import BuildSlaveTooOldError
+
 
 class Bzr(Source):
 
     name = 'bzr'
-    renderables = [ 'repourl', 'baseURL' ]
+    renderables = ['repourl', 'baseURL']
 
     def __init__(self, repourl=None, baseURL=None, mode='incremental',
                  method=None, defaultBranch=None, **kwargs):
@@ -41,13 +47,15 @@ class Bzr(Source):
                              " baseURL")
 
         if repourl is None and baseURL is None:
-            raise ValueError("you must privide at least one of repourl and"
+            raise ValueError("you must provide at least one of repourl and"
                              " baseURL")
 
         if baseURL is not None and defaultBranch is None:
             raise ValueError("you must provide defaultBranch with baseURL")
 
-        assert self.mode in ['incremental', 'full']
+        if not self._hasAttrGroupMember('mode', self.mode):
+            raise ValueError("mode %s is not one of %s" %
+                             (self.mode, self._listAttrGroupMembers('mode')))
 
         if self.mode == 'full':
             assert self.method in ['clean', 'fresh', 'clobber', 'copy', None]
@@ -63,40 +71,43 @@ class Bzr(Source):
             self.repourl = os.path.join(self.baseURL, self.branch)
 
         d = self.checkBzr()
+
+        @d.addCallback
         def checkInstall(bzrInstalled):
             if not bzrInstalled:
-                raise BuildSlaveTooOldError("bzr is not installed on slave")
+                raise WorkerTooOldError("bzr is not installed on worker")
             return 0
 
-        d.addCallback(checkInstall)
-        if self.mode == 'full':
-            d.addCallback(lambda _: self.full())
-        elif self.mode == 'incremental':
-            d.addCallback(lambda _: self.incremental())
+        d.addCallback(lambda _: self.sourcedirIsPatched())
 
+        @d.addCallback
+        def checkPatched(patched):
+            if patched:
+                return self._dovccmd(['clean-tree', '--ignored', '--force'])
+            return 0
+
+        d.addCallback(self._getAttrGroupMember('mode', self.mode))
+
+        if patch:
+            d.addCallback(self.patch, patch)
         d.addCallback(self.parseGotRevision)
         d.addCallback(self.finish)
         d.addErrback(self.failed)
         return d
 
-    def incremental(self):
-        d = self._sourcedirIsUpdatable()
-        def _cmd(updatable):
-            if updatable:
-                command = ['update']
-            else:
-                command = ['checkout', self.repourl, '.']
-
+    @defer.inlineCallbacks
+    def mode_incremental(self, _):
+        updatable = yield self._sourcedirIsUpdatable()
+        if updatable:
+            command = ['update']
             if self.revision:
                 command.extend(['-r', self.revision])
-            return command
-
-        d.addCallback(_cmd)
-        d.addCallback(self._dovccmd)
-        return d
+            yield self._dovccmd(command)
+        else:
+            yield self._doFull()
 
     @defer.inlineCallbacks
-    def full(self):
+    def mode_full(self, _):
         if self.method == 'clobber':
             yield self.clobber()
             return
@@ -116,34 +127,40 @@ class Bzr(Source):
         else:
             raise ValueError("Unknown method, check your configuration")
 
-    def clobber(self):
-        cmd = buildstep.RemoteCommand('rmdir', {'dir': self.workdir,
-                                                'logEnviron': self.logEnviron,})
+    def _clobber(self):
+        cmd = remotecommand.RemoteCommand('rmdir', {'dir': self.workdir,
+                                                    'logEnviron': self.logEnviron, })
         cmd.useLog(self.stdio_log, False)
         d = self.runCommand(cmd)
+
+        @d.addCallback
         def checkRemoval(res):
-            if res != 0:
+            if cmd.rc != 0:
                 raise RuntimeError("Failed to delete directory")
-            return res
-        d.addCallback(lambda _: checkRemoval(cmd.rc))
+            return cmd.rc
+        return d
+
+    def clobber(self):
+        d = self._clobber()
         d.addCallback(lambda _: self._doFull())
         return d
 
     def copy(self):
-        cmd = buildstep.RemoteCommand('rmdir', {'dir': 'build',
-                                                'logEnviron': self.logEnviron,})
+        cmd = remotecommand.RemoteCommand('rmdir', {'dir': 'build',
+                                                    'logEnviron': self.logEnviron, })
         cmd.useLog(self.stdio_log, False)
         d = self.runCommand(cmd)
-        d.addCallback(lambda _: self.incremental())
+        d.addCallback(self.mode_incremental)
+
+        @d.addCallback
         def copy(_):
-            cmd = buildstep.RemoteCommand('cpdir',
-                                          {'fromdir': 'source',
-                                           'todir':'build',
-                                           'logEnviron': self.logEnviron,})
+            cmd = remotecommand.RemoteCommand('cpdir',
+                                              {'fromdir': 'source',
+                                               'todir': 'build',
+                                               'logEnviron': self.logEnviron, })
             cmd.useLog(self.stdio_log, False)
             d = self.runCommand(cmd)
             return d
-        d.addCallback(copy)
         return d
 
     def clean(self):
@@ -166,18 +183,42 @@ class Bzr(Source):
         command = ['checkout', self.repourl, '.']
         if self.revision:
             command.extend(['-r', self.revision])
-        d = self._dovccmd(command)
+
+        if self.retry:
+            abandonOnFailure = (self.retry[1] <= 0)
+        else:
+            abandonOnFailure = True
+        d = self._dovccmd(command, abandonOnFailure=abandonOnFailure)
+
+        def _retry(res):
+            if self.stopped or res == 0:
+                return res
+            delay, repeats = self.retry
+            if repeats > 0:
+                log.msg("Checkout failed, trying %d more times after %d seconds"
+                        % (repeats, delay))
+                self.retry = (delay, repeats - 1)
+                df = defer.Deferred()
+                df.addCallback(lambda _: self._clobber())
+                df.addCallback(lambda _: self._doFull())
+                reactor.callLater(delay, df.callback, None)
+                return df
+            return res
+
+        if self.retry:
+            d.addCallback(_retry)
         return d
 
     def finish(self, res):
         d = defer.succeed(res)
+
+        @d.addCallback
         def _gotResults(results):
             self.setStatus(self.cmd, results)
-            log.msg("Closing log, sending result of the command %s " % \
+            log.msg("Closing log, sending result of the command %s " %
                     (self.cmd))
             return results
-        d.addCallback(_gotResults)
-        d.addCallbacks(self.finished, self.checkDisconnect)
+        d.addCallback(self.finished)
         return d
 
     def _sourcedirIsUpdatable(self):
@@ -190,31 +231,30 @@ class Bzr(Source):
         return lastChange
 
     def _dovccmd(self, command, abandonOnFailure=True, collectStdout=False):
-        cmd = buildstep.RemoteShellCommand(self.workdir, ['bzr'] + command,
-                                           env=self.env,
-                                           logEnviron=self.logEnviron,
-                                           timeout=self.timeout,
-                                           collectStdout=collectStdout)
+        cmd = remotecommand.RemoteShellCommand(self.workdir, ['bzr'] + command,
+                                               env=self.env,
+                                               logEnviron=self.logEnviron,
+                                               timeout=self.timeout,
+                                               collectStdout=collectStdout)
         cmd.useLog(self.stdio_log, False)
         d = self.runCommand(cmd)
-        def evaluateCommand(cmd):
+
+        @d.addCallback
+        def evaluateCommand(_):
             if abandonOnFailure and cmd.didFail():
                 log.msg("Source step failed while running command %s" % cmd)
                 raise buildstep.BuildStepFailed()
             if collectStdout:
                 return cmd.stdout
-            else:
-                return cmd.rc
-        d.addCallback(lambda _: evaluateCommand(cmd))
+            return cmd.rc
         return d
 
     def checkBzr(self):
         d = self._dovccmd(['--version'])
+
+        @d.addCallback
         def check(res):
-            if res == 0:
-                return True
-            return False
-        d.addCallback(check)
+            return res == 0
         return d
 
     def _getMethod(self):
@@ -228,6 +268,8 @@ class Bzr(Source):
     def parseGotRevision(self, _):
         d = self._dovccmd(["version-info", "--custom", "--template='{revno}"],
                           collectStdout=True)
+
+        @d.addCallback
         def setrev(stdout):
             revision = stdout.strip("'")
             try:
@@ -239,6 +281,4 @@ class Bzr(Source):
             log.msg("Got Git revision %s" % (revision, ))
             self.updateSourceProperty('got_revision', revision)
             return 0
-        d.addCallback(setrev)
         return d
-
